@@ -6,18 +6,18 @@ from typing import Any, Dict, List, Optional, Text, Tuple, Type
 
 import numpy as np
 
+import rasa.shared.utils.io
 import rasa.utils.io as io_utils
-from rasa.constants import DOCS_URL_TRAINING_DATA_NLU
+from rasa.shared.constants import DOCS_URL_TRAINING_DATA_NLU
 from rasa.nlu.classifiers import LABEL_RANKING_LENGTH
 from rasa.nlu.featurizers.featurizer import DenseFeaturizer
 from rasa.nlu.components import Component
 from rasa.nlu.classifiers.classifier import IntentClassifier
 from rasa.nlu.config import RasaNLUModelConfig
-from rasa.nlu.constants import DENSE_FEATURE_NAMES, TEXT
-from rasa.nlu.featurizers.featurizer import sequence_to_sentence_features
+from rasa.shared.nlu.constants import TEXT
 from rasa.nlu.model import Metadata
-from rasa.nlu.training_data import Message, TrainingData
-import rasa.utils.common as common_utils
+from rasa.shared.nlu.training_data.training_data import TrainingData
+from rasa.shared.nlu.training_data.message import Message
 
 logger = logging.getLogger(__name__)
 
@@ -96,40 +96,49 @@ class SklearnIntentClassifier(IntentClassifier):
         labels = [e.get("intent") for e in training_data.intent_examples]
 
         if len(set(labels)) < 2:
-            common_utils.raise_warning(
+            rasa.shared.utils.io.raise_warning(
                 "Can not train an intent classifier as there are not "
                 "enough intents. Need at least 2 different intents. "
                 "Skipping training of intent classifier.",
                 docs=DOCS_URL_TRAINING_DATA_NLU,
             )
-        else:
-            y = self.transform_labels_str2num(labels)
-            X = np.stack(
-                [
-                    sequence_to_sentence_features(
-                        example.get(DENSE_FEATURE_NAMES[TEXT])
-                    )
-                    for example in training_data.intent_examples
-                ]
-            )
-            # reduce dimensionality
-            X = np.reshape(X, (len(X), -1))
+            return
 
-            self.clf = self._create_classifier(num_threads, y)
+        y = self.transform_labels_str2num(labels)
+        X = np.stack(
+            [
+                self._get_sentence_features(example)
+                for example in training_data.intent_examples
+            ]
+        )
+        # reduce dimensionality
+        X = np.reshape(X, (len(X), -1))
 
-            with warnings.catch_warnings():
-                # sklearn raises lots of
-                # "UndefinedMetricWarning: F - score is ill - defined"
-                # if there are few intent examples, this is needed to prevent it
-                warnings.simplefilter("ignore")
-                self.clf.fit(X, y)
+        self.clf = self._create_classifier(num_threads, y)
 
-    def _num_cv_splits(self, y) -> int:
+        with warnings.catch_warnings():
+            # sklearn raises lots of
+            # "UndefinedMetricWarning: F - score is ill - defined"
+            # if there are few intent examples, this is needed to prevent it
+            warnings.simplefilter("ignore")
+            self.clf.fit(X, y)
+
+    @staticmethod
+    def _get_sentence_features(message: Message) -> np.ndarray:
+        _, sentence_features = message.get_dense_features(TEXT)
+        if sentence_features is not None:
+            return sentence_features.features[0]
+
+        raise ValueError(
+            "No sentence features present. Not able to train sklearn policy."
+        )
+
+    def _num_cv_splits(self, y: np.ndarray) -> int:
         folds = self.component_config["max_cross_validation_folds"]
         return max(2, min(folds, np.min(np.bincount(y)) // 5))
 
     def _create_classifier(
-        self, num_threads: int, y
+        self, num_threads: int, y: np.ndarray
     ) -> "sklearn.model_selection.GridSearchCV":
         from sklearn.model_selection import GridSearchCV
         from sklearn.svm import SVC
@@ -166,9 +175,8 @@ class SklearnIntentClassifier(IntentClassifier):
             intent = None
             intent_ranking = []
         else:
-            X = sequence_to_sentence_features(
-                message.get(DENSE_FEATURE_NAMES[TEXT])
-            ).reshape(1, -1)
+            X = self._get_sentence_features(message).reshape(1, -1)
+
             intent_ids, probabilities = self.predict(X)
             intents = self.transform_labels_num2str(np.ravel(intent_ids))
             # `predict` returns a matrix as it is supposed
