@@ -1,11 +1,13 @@
 import asyncio
-import functools
+import concurrent.futures
 import logging
 import multiprocessing
 import os
 import tempfile
 import traceback
+from collections import defaultdict
 from functools import reduce, wraps
+from http import HTTPStatus
 from inspect import isawaitable
 from pathlib import Path
 from typing import (
@@ -18,8 +20,10 @@ from typing import (
     Dict,
     TYPE_CHECKING,
     NoReturn,
+    Coroutine,
 )
 
+import aiohttp
 from sanic import Sanic, response
 from sanic.request import Request
 from sanic.response import HTTPResponse
@@ -35,6 +39,7 @@ import rasa.utils.io
 from rasa.shared.core.training_data.story_writer.yaml_story_writer import (
     YAMLStoryWriter,
 )
+from rasa.shared.importers.importer import TrainingDataImporter
 from rasa.shared.nlu.training_data.formats import RasaYAMLReader
 from rasa import model
 from rasa.constants import DEFAULT_RESPONSE_TIMEOUT, MINIMUM_COMPATIBLE_VERSION
@@ -61,7 +66,7 @@ from rasa.core.tracker_store import TrackerStore
 from rasa.shared.core.trackers import DialogueStateTracker, EventVerbosity
 from rasa.core.utils import AvailableEndpoints
 from rasa.nlu.emulators.no_emulator import NoEmulator
-from rasa.nlu.test import run_evaluation
+from rasa.nlu.test import run_evaluation, CVEvaluationResult
 from rasa.utils.endpoints import EndpointConfig
 
 if TYPE_CHECKING:
@@ -79,14 +84,25 @@ EXECUTE_SIDE_EFFECTS_QUERY_KEY = "execute_side_effects"
 
 
 class ErrorResponse(Exception):
+    """Common exception to handle failing API requests."""
+
     def __init__(
         self,
-        status: int,
+        status: Union[int, HTTPStatus],
         reason: Text,
         message: Text,
         details: Any = None,
         help_url: Optional[Text] = None,
     ) -> None:
+        """Creates error.
+
+        Args:
+            status: The HTTP status code to return.
+            reason: Short summary of the error.
+            message: Detailed explanation of the error.
+            details: Additional details which describe the error. Must be serializable.
+            help_url: URL where users can get further help (e.g. docs).
+        """
         self.error_info = {
             "version": rasa.__version__,
             "status": "failure",
@@ -123,7 +139,7 @@ def ensure_loaded_agent(app: Sanic, require_core_is_ready=False):
                 else app.agent.is_ready()
             ):
                 raise ErrorResponse(
-                    409,
+                    HTTPStatus.CONFLICT,
                     "Conflict",
                     "No agent loaded. To continue processing, a "
                     "model of a trained agent needs to be loaded.",
@@ -189,7 +205,7 @@ def requires_auth(app: Sanic, token: Optional[Text] = None) -> Callable[[Any], A
                         result = await result
                     return result
                 raise ErrorResponse(
-                    403,
+                    HTTPStatus.FORBIDDEN,
                     "NotAuthorized",
                     "User has insufficient permissions.",
                     help_url=_docs(
@@ -203,7 +219,7 @@ def requires_auth(app: Sanic, token: Optional[Text] = None) -> Callable[[Any], A
                     result = await result
                 return result
             raise ErrorResponse(
-                401,
+                HTTPStatus.UNAUTHORIZED,
                 "NotAuthenticated",
                 "User is not authenticated.",
                 help_url=_docs(
@@ -228,7 +244,7 @@ def event_verbosity_parameter(
     except KeyError:
         enum_values = ", ".join([e.name for e in EventVerbosity])
         raise ErrorResponse(
-            400,
+            HTTPStatus.BAD_REQUEST,
             "BadRequest",
             "Invalid parameter value for 'include_events'. "
             "Should be one of {}".format(enum_values),
@@ -316,7 +332,7 @@ async def update_conversation_with_events(
 def validate_request_body(request: Request, error_message: Text) -> None:
     """Check if `request` has a body."""
     if not request.body:
-        raise ErrorResponse(400, "BadRequest", error_message)
+        raise ErrorResponse(HTTPStatus.BAD_REQUEST, "BadRequest", error_message)
 
 
 async def authenticate(_: Request) -> NoReturn:
@@ -381,7 +397,7 @@ def _create_emulator(mode: Optional[Text]) -> NoEmulator:
         return DialogflowEmulator()
     else:
         raise ErrorResponse(
-            400,
+            HTTPStatus.BAD_REQUEST,
             "BadRequest",
             "Invalid parameter value for 'emulation_mode'. "
             "Should be one of 'WIT', 'LUIS', 'DIALOGFLOW'.",
@@ -423,12 +439,14 @@ async def _load_agent(
     except Exception as e:
         logger.debug(traceback.format_exc())
         raise ErrorResponse(
-            500, "LoadingError", f"An unexpected error occurred. Error: {e}"
+            HTTPStatus.INTERNAL_SERVER_ERROR,
+            "LoadingError",
+            f"An unexpected error occurred. Error: {e}",
         )
 
     if not loaded_agent:
         raise ErrorResponse(
-            400,
+            HTTPStatus.BAD_REQUEST,
             "BadRequest",
             f"Agent with name '{model_path}' could not be loaded.",
             {"parameter": "model", "in": "query"},
@@ -462,6 +480,124 @@ def add_root_route(app: Sanic):
         return response.text("Hello from Rasa: " + rasa.__version__)
 
 
+def async_if_callback_url(f: Callable[..., Coroutine]) -> Callable:
+    """Decorator to enable async request handling.
+
+    If the incoming HTTP request specified a `callback_url` query parameter, the request
+    will return immediately with a 204 while the actual request response will
+    be sent to the `callback_url`. If an error happens, the error payload will also
+    be sent to the `callback_url`.
+
+    Args:
+        f: The request handler function which should be decorated.
+
+    Returns:
+        The decorated function.
+    """
+
+    @wraps(f)
+    async def decorated_function(
+        request: Request, *args: Any, **kwargs: Any
+    ) -> HTTPResponse:
+        callback_url = request.args.get("callback_url")
+        # Only process request asynchronously if the user specified a `callback_url`
+        # query parameter.
+        if not callback_url:
+            return await f(request, *args, **kwargs)
+
+        async def wrapped() -> None:
+            try:
+                result: HTTPResponse = await f(request, *args, **kwargs)
+                payload = dict(
+                    data=result.body, headers={"Content-Type": result.content_type}
+                )
+                logger.debug(
+                    "Asynchronous processing of request was successful. "
+                    "Sending result to callback URL."
+                )
+
+            except Exception as e:
+                if not isinstance(e, ErrorResponse):
+                    logger.error(e)
+                    e = ErrorResponse(
+                        HTTPStatus.INTERNAL_SERVER_ERROR,
+                        "UnexpectedError",
+                        f"An unexpected error occurred. Error: {e}",
+                    )
+                # If an error happens, we send the error payload to the `callback_url`
+                payload = dict(json=e.error_info)
+                logger.debug(
+                    "Error happened when processing request asynchronously. "
+                    "Sending error to callback URL."
+                )
+            async with aiohttp.ClientSession() as session:
+                await session.post(callback_url, raise_for_status=True, **payload)
+
+        # Run the request in the background on the event loop
+        request.app.add_task(wrapped())
+
+        # The incoming request will return immediately with a 204
+        return response.empty()
+
+    return decorated_function
+
+
+def run_in_thread(f: Callable[..., Coroutine]) -> Callable:
+    """Decorator which runs request on a separate thread.
+
+    Some requests (e.g. training or cross-validation) are computional intense requests.
+    This means that they will block the event loop and hence the processing of other
+    requests. This decorator can be used to process these requests on a separate thread
+    to avoid blocking the processing of incoming requests.
+
+    Args:
+        f: The request handler function which should be decorated.
+
+    Returns:
+        The decorated function.
+    """
+
+    @wraps(f)
+    async def decorated_function(
+        request: Request, *args: Any, **kwargs: Any
+    ) -> HTTPResponse:
+        # Use a sync wrapper for our `async` function as `run_in_executor` only supports
+        # sync functions
+        def run() -> HTTPResponse:
+            # This is a new thread, so we need to create and set a new event loop
+            thread_loop = asyncio.new_event_loop()
+            asyncio.set_event_loop(thread_loop)
+
+            try:
+                return thread_loop.run_until_complete(f(request, *args, **kwargs))
+            finally:
+                thread_loop.close()
+
+        with concurrent.futures.ThreadPoolExecutor() as pool:
+            return await request.app.loop.run_in_executor(pool, run)
+
+    return decorated_function
+
+
+def inject_temp_dir(f: Callable[..., Coroutine]) -> Callable:
+    """Decorator to inject a temporary directory before a request and clean up after.
+
+    Args:
+        f: The request handler function which should be decorated.
+
+    Returns:
+        The decorated function.
+    """
+
+    @wraps(f)
+    async def decorated_function(*args: Any, **kwargs: Any) -> HTTPResponse:
+        with tempfile.TemporaryDirectory() as directory:
+            # Decorated request handles need to have a parameter `temporary_directory`
+            return await f(*args, temporary_directory=Path(directory), **kwargs)
+
+    return decorated_function
+
+
 def create_app(
     agent: Optional["Agent"] = None,
     cors_origins: Union[Text, List[Text], None] = "*",
@@ -550,7 +686,9 @@ def create_app(
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500, "ConversationError", f"An unexpected error occurred. Error: {e}"
+                HTTPStatus.INTERNAL_SERVER_ERROR,
+                "ConversationError",
+                f"An unexpected error occurred. Error: {e}",
             )
 
     @app.post("/conversations/<conversation_id:path>/tracker/events")
@@ -590,7 +728,9 @@ def create_app(
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500, "ConversationError", f"An unexpected error occurred. Error: {e}"
+                HTTPStatus.INTERNAL_SERVER_ERROR,
+                "ConversationError",
+                f"An unexpected error occurred. Error: {e}",
             )
 
     def _get_events_from_request_body(request: Request) -> List[Event]:
@@ -608,7 +748,7 @@ def create_app(
                 f"Request JSON: {request.json}"
             )
             raise ErrorResponse(
-                400,
+                HTTPStatus.BAD_REQUEST,
                 "BadRequest",
                 "Couldn't extract a proper event from the request body.",
                 {"parameter": "", "in": "body"},
@@ -642,7 +782,9 @@ def create_app(
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500, "ConversationError", f"An unexpected error occurred. Error: {e}"
+                HTTPStatus.INTERNAL_SERVER_ERROR,
+                "ConversationError",
+                f"An unexpected error occurred. Error: {e}",
             )
 
     @app.get("/conversations/<conversation_id:path>/story")
@@ -666,7 +808,9 @@ def create_app(
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500, "ConversationError", f"An unexpected error occurred. Error: {e}"
+                HTTPStatus.INTERNAL_SERVER_ERROR,
+                "ConversationError",
+                f"An unexpected error occurred. Error: {e}",
             )
 
     @app.post("/conversations/<conversation_id:path>/execute")
@@ -679,7 +823,7 @@ def create_app(
 
         if not action_to_execute:
             raise ErrorResponse(
-                400,
+                HTTPStatus.BAD_REQUEST,
                 "BadRequest",
                 "Name of the action not provided in request body.",
                 {"parameter": "name", "in": "body"},
@@ -707,7 +851,9 @@ def create_app(
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500, "ConversationError", f"An unexpected error occurred. Error: {e}"
+                HTTPStatus.INTERNAL_SERVER_ERROR,
+                "ConversationError",
+                f"An unexpected error occurred. Error: {e}",
             )
 
         state = tracker.current_state(verbosity)
@@ -730,7 +876,7 @@ def create_app(
 
         if not intent_to_trigger:
             raise ErrorResponse(
-                400,
+                HTTPStatus.BAD_REQUEST,
                 "BadRequest",
                 "Name of the intent not provided in request body.",
                 {"parameter": "name", "in": "body"},
@@ -746,7 +892,7 @@ def create_app(
                 output_channel = _get_output_channel(request, tracker)
                 if intent_to_trigger not in app.agent.domain.intents:
                     raise ErrorResponse(
-                        404,
+                        HTTPStatus.NOT_FOUND,
                         "NotFound",
                         f"The intent {trigger_intent} does not exist in the domain.",
                     )
@@ -761,7 +907,9 @@ def create_app(
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500, "ConversationError", f"An unexpected error occurred. Error: {e}"
+                HTTPStatus.INTERNAL_SERVER_ERROR,
+                "ConversationError",
+                f"An unexpected error occurred. Error: {e}",
             )
 
         state = tracker.current_state(verbosity)
@@ -787,7 +935,9 @@ def create_app(
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500, "ConversationError", f"An unexpected error occurred. Error: {e}"
+                HTTPStatus.INTERNAL_SERVER_ERROR,
+                "ConversationError",
+                f"An unexpected error occurred. Error: {e}",
             )
 
     @app.post("/conversations/<conversation_id:path>/messages")
@@ -811,7 +961,7 @@ def create_app(
         # TODO: implement for agent / bot
         if sender != "user":
             raise ErrorResponse(
-                400,
+                HTTPStatus.BAD_REQUEST,
                 "BadRequest",
                 "Currently, only user messages can be passed to this endpoint. "
                 "Messages of sender '{}' cannot be handled.".format(sender),
@@ -828,14 +978,17 @@ def create_app(
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500, "ConversationError", f"An unexpected error occurred. Error: {e}"
+                HTTPStatus.INTERNAL_SERVER_ERROR,
+                "ConversationError",
+                f"An unexpected error occurred. Error: {e}",
             )
 
     @app.post("/model/train")
     @requires_auth(app, auth_token)
-    async def train(request: Request) -> HTTPResponse:
-        """Train a Rasa Model."""
-
+    @async_if_callback_url
+    @run_in_thread
+    @inject_temp_dir
+    async def train(request: Request, temporary_directory: Path) -> HTTPResponse:
         validate_request_body(
             request,
             "You must provide training data in the request body in order to "
@@ -843,32 +996,30 @@ def create_app(
         )
 
         if request.headers.get("Content-type") == YAML_CONTENT_TYPE:
-            training_payload = _training_payload_from_yaml(request)
+            training_payload = _training_payload_from_yaml(request, temporary_directory)
         else:
-            training_payload = _training_payload_from_json(request)
+            training_payload = _training_payload_from_json(request, temporary_directory)
 
         try:
             with app.active_training_processes.get_lock():
                 app.active_training_processes.value += 1
 
-            loop = asyncio.get_event_loop()
-
-            from rasa import train as train_model
+            from rasa.train import train_async
 
             # pass `None` to run in default executor
-            model_path = await loop.run_in_executor(
-                None, functools.partial(train_model, **training_payload)
-            )
+            training_result = await train_async(**training_payload)
 
-            if model_path:
-                filename = os.path.basename(model_path)
+            if training_result.model:
+                filename = os.path.basename(training_result.model)
 
                 return await response.file(
-                    model_path, filename=filename, headers={"filename": filename}
+                    training_result.model,
+                    filename=filename,
+                    headers={"filename": filename},
                 )
             else:
                 raise ErrorResponse(
-                    500,
+                    HTTPStatus.INTERNAL_SERVER_ERROR,
                     "TrainingError",
                     "Ran training, but it finished without a trained model.",
                 )
@@ -876,14 +1027,14 @@ def create_app(
             raise e
         except InvalidDomain as e:
             raise ErrorResponse(
-                400,
+                HTTPStatus.BAD_REQUEST,
                 "InvalidDomainError",
                 f"Provided domain file is invalid. Error: {e}",
             )
         except Exception as e:
             logger.error(traceback.format_exc())
             raise ErrorResponse(
-                500,
+                HTTPStatus.INTERNAL_SERVER_ERROR,
                 "TrainingError",
                 f"An unexpected error occurred during training. Error: {e}",
             )
@@ -894,7 +1045,10 @@ def create_app(
     @app.post("/model/test/stories")
     @requires_auth(app, auth_token)
     @ensure_loaded_agent(app, require_core_is_ready=True)
-    async def evaluate_stories(request: Request) -> HTTPResponse:
+    @inject_temp_dir
+    async def evaluate_stories(
+        request: Request, temporary_directory: Path
+    ) -> HTTPResponse:
         """Evaluate stories against the currently loaded model."""
         validate_request_body(
             request,
@@ -902,7 +1056,7 @@ def create_app(
             "evaluate your model.",
         )
 
-        test_data = _test_data_file_from_payload(request)
+        test_data = _test_data_file_from_payload(request, temporary_directory)
 
         use_e2e = rasa.utils.endpoints.bool_arg(request, "e2e", default=False)
 
@@ -914,14 +1068,19 @@ def create_app(
         except Exception as e:
             logger.error(traceback.format_exc())
             raise ErrorResponse(
-                500,
+                HTTPStatus.INTERNAL_SERVER_ERROR,
                 "TestingError",
                 f"An unexpected error occurred during evaluation. Error: {e}",
             )
 
     @app.post("/model/test/intents")
     @requires_auth(app, auth_token)
-    async def evaluate_intents(request: Request) -> HTTPResponse:
+    @async_if_callback_url
+    @run_in_thread
+    @inject_temp_dir
+    async def evaluate_intents(
+        request: Request, temporary_directory: Path
+    ) -> HTTPResponse:
         """Evaluate intents against a Rasa model."""
         validate_request_body(
             request,
@@ -929,11 +1088,49 @@ def create_app(
             "evaluate your model.",
         )
 
-        test_data = _test_data_file_from_payload(request)
+        cross_validation_folds = request.args.get("cross_validation_folds")
+        is_yaml_payload = request.headers.get("Content-type") == YAML_CONTENT_TYPE
+        test_coroutine = None
 
+        if is_yaml_payload:
+            payload = _training_payload_from_yaml(request, temporary_directory)
+            config_file = payload.get("config")
+            test_data = payload.get("training_files")
+
+            if cross_validation_folds:
+                test_coroutine = _cross_validate(
+                    test_data, config_file, int(cross_validation_folds)
+                )
+        else:
+            test_data = _test_data_file_from_payload(request, temporary_directory)
+            if cross_validation_folds:
+                raise ErrorResponse(
+                    HTTPStatus.BAD_REQUEST,
+                    "TestingError",
+                    "Cross-validation is only supported for YAML data.",
+                )
+
+        if not cross_validation_folds:
+            test_coroutine = _evaluate_model_using_test_set(
+                request.args.get("model"), test_data
+            )
+
+        try:
+            evaluation = await test_coroutine
+            return response.json(evaluation)
+        except Exception as e:
+            logger.error(traceback.format_exc())
+            raise ErrorResponse(
+                HTTPStatus.INTERNAL_SERVER_ERROR,
+                "TestingError",
+                f"An unexpected error occurred during evaluation. Error: {e}",
+            )
+
+    async def _evaluate_model_using_test_set(
+        model_path: Text, test_data_file: Text
+    ) -> Dict:
         eval_agent = app.agent
 
-        model_path = request.args.get("model", None)
         if model_path:
             model_server = app.agent.model_server
             if model_server is not None:
@@ -942,26 +1139,62 @@ def create_app(
                 model_path, model_server, app.agent.remote_storage
             )
 
-        data_path = os.path.abspath(test_data)
+        data_path = os.path.abspath(test_data_file)
 
         if not eval_agent.model_directory or not os.path.exists(
             eval_agent.model_directory
         ):
-            raise ErrorResponse(409, "Conflict", "Loaded model file not found.")
+            raise ErrorResponse(
+                HTTPStatus.CONFLICT, "Conflict", "Loaded model file not found."
+            )
 
         model_directory = eval_agent.model_directory
         _, nlu_model = model.get_model_subdirectories(model_directory)
 
-        try:
-            evaluation = run_evaluation(data_path, nlu_model, disable_plotting=True)
-            return response.json(evaluation)
-        except Exception as e:
-            logger.error(traceback.format_exc())
-            raise ErrorResponse(
-                500,
-                "TestingError",
-                f"An unexpected error occurred during evaluation. Error: {e}",
-            )
+        return run_evaluation(
+            data_path, nlu_model, disable_plotting=True, report_as_dict=True
+        )
+
+    async def _cross_validate(data_file: Text, config_file: Text, folds: int) -> Dict:
+        importer = TrainingDataImporter.load_from_dict(
+            config=None, config_path=config_file, training_data_paths=[data_file]
+        )
+        config = await importer.get_config()
+        nlu_data = await importer.get_nlu_data()
+
+        evaluations = rasa.nlu.cross_validate(
+            data=nlu_data,
+            n_folds=folds,
+            nlu_config=config,
+            disable_plotting=True,
+            errors=True,
+            report_as_dict=True,
+        )
+        evaluation_results = _get_evaluation_results(*evaluations)
+
+        return evaluation_results
+
+    def _get_evaluation_results(
+        intent_report: CVEvaluationResult,
+        entity_report: CVEvaluationResult,
+        response_selector_report: CVEvaluationResult,
+    ) -> Dict[Text, Any]:
+        eval_name_mapping = {
+            "intent_evaluation": intent_report,
+            "entity_evaluation": entity_report,
+            "response_selection_evaluation": response_selector_report,
+        }
+
+        result = defaultdict(dict)
+        for evaluation_name, evaluation in eval_name_mapping.items():
+            report = evaluation.evaluation.get("report", {})
+            averages = report.get("weighted avg", {})
+            result[evaluation_name]["report"] = report
+            result[evaluation_name]["precision"] = averages.get("precision")
+            result[evaluation_name]["f1_score"] = averages.get("1-score")
+            result[evaluation_name]["errors"] = evaluation.evaluation.get("errors", [])
+
+        return result
 
     @app.post("/model/predict")
     @requires_auth(app, auth_token)
@@ -983,7 +1216,7 @@ def create_app(
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                400,
+                HTTPStatus.BAD_REQUEST,
                 "BadRequest",
                 f"Supplied events are not valid. {e}",
                 {"parameter": "", "in": "body"},
@@ -998,7 +1231,9 @@ def create_app(
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500, "PredictionError", f"An unexpected error occurred. Error: {e}"
+                HTTPStatus.INTERNAL_SERVER_ERROR,
+                "PredictionError",
+                f"An unexpected error occurred. Error: {e}",
             )
 
     @app.post("/model/parse")
@@ -1022,7 +1257,9 @@ def create_app(
             except Exception as e:
                 logger.debug(traceback.format_exc())
                 raise ErrorResponse(
-                    400, "ParsingError", f"An unexpected error occurred. Error: {e}"
+                    HTTPStatus.BAD_REQUEST,
+                    "ParsingError",
+                    f"An unexpected error occurred. Error: {e}",
                 )
             response_data = emulator.normalise_response_json(parsed_data)
 
@@ -1031,7 +1268,9 @@ def create_app(
         except Exception as e:
             logger.debug(traceback.format_exc())
             raise ErrorResponse(
-                500, "ParsingError", f"An unexpected error occurred. Error: {e}"
+                HTTPStatus.INTERNAL_SERVER_ERROR,
+                "ParsingError",
+                f"An unexpected error occurred. Error: {e}",
             )
 
     @app.put("/model")
@@ -1049,7 +1288,7 @@ def create_app(
             except TypeError as e:
                 logger.debug(traceback.format_exc())
                 raise ErrorResponse(
-                    400,
+                    HTTPStatus.BAD_REQUEST,
                     "BadRequest",
                     f"Supplied 'model_server' is not valid. Error: {e}",
                     {"parameter": "model_server", "in": "body"},
@@ -1060,7 +1299,7 @@ def create_app(
         )
 
         logger.debug(f"Successfully loaded model '{model_path}'.")
-        return response.json(None, status=204)
+        return response.json(None, status=HTTPStatus.NO_CONTENT)
 
     @app.delete("/model")
     @requires_auth(app, auth_token)
@@ -1070,14 +1309,13 @@ def create_app(
         app.agent = Agent(lock_store=app.agent.lock_store)
 
         logger.debug(f"Successfully unloaded model '{model_file}'.")
-        return response.json(None, status=204)
+        return response.json(None, status=HTTPStatus.NO_CONTENT)
 
     @app.get("/domain")
     @requires_auth(app, auth_token)
     @ensure_loaded_agent(app)
     async def get_domain(request: Request) -> HTTPResponse:
         """Get current domain in yaml or json format."""
-
         accepts = request.headers.get("Accept", default=JSON_CONTENT_TYPE)
         if accepts.endswith("json"):
             domain = app.agent.domain.as_dict()
@@ -1085,11 +1323,11 @@ def create_app(
         elif accepts.endswith("yml") or accepts.endswith("yaml"):
             domain_yaml = app.agent.domain.as_yaml()
             return response.text(
-                domain_yaml, status=200, content_type=YAML_CONTENT_TYPE
+                domain_yaml, status=HTTPStatus.OK, content_type=YAML_CONTENT_TYPE
             )
         else:
             raise ErrorResponse(
-                406,
+                HTTPStatus.NOT_ACCEPTABLE,
                 "NotAcceptable",
                 f"Invalid Accept header. Domain can be "
                 f"provided as "
@@ -1142,16 +1380,20 @@ def _get_output_channel(
     )
 
 
-def _test_data_file_from_payload(request: Request) -> Text:
+def _test_data_file_from_payload(request: Request, temporary_directory: Path) -> Text:
     if request.headers.get("Content-type") == YAML_CONTENT_TYPE:
-        return str(_training_payload_from_yaml(request)["training_files"])
+        return str(
+            _training_payload_from_yaml(request, temporary_directory)["training_files"]
+        )
     else:
         return rasa.utils.io.create_temporary_file(
             request.body, mode="w+b", suffix=".md"
         )
 
 
-def _training_payload_from_json(request: Request) -> Dict[Text, Union[Text, bool]]:
+def _training_payload_from_json(
+    request: Request, temp_dir: Path
+) -> Dict[Text, Union[Text, bool]]:
     logger.debug(
         "Extracting JSON payload with Markdown training data from request body."
     )
@@ -1159,10 +1401,6 @@ def _training_payload_from_json(request: Request) -> Dict[Text, Union[Text, bool
     request_payload = request.json
     _validate_json_training_payload(request_payload)
 
-    # create a temporary directory to store config, domain and
-    # training data
-    temp_dir = tempfile.mkdtemp()
-
     config_path = os.path.join(temp_dir, "config.yml")
 
     rasa.shared.utils.io.write_text_file(request_payload["config"], config_path)
@@ -1186,17 +1424,17 @@ def _training_payload_from_json(request: Request) -> Dict[Text, Union[Text, bool
         domain_path = os.path.join(temp_dir, "domain.yml")
         rasa.shared.utils.io.write_text_file(request_payload["domain"], domain_path)
 
-    model_output_directory = _model_output_directory(
-        request_payload.get(
-            "save_to_default_model_directory",
-            request.args.get("save_to_default_model_directory", True),
-        )
-    )
+    model_output_directory = str(temp_dir)
+    if request_payload.get(
+        "save_to_default_model_directory",
+        request.args.get("save_to_default_model_directory", True),
+    ):
+        model_output_directory = DEFAULT_MODELS_PATH
 
     return dict(
         domain=domain_path,
         config=config_path,
-        training_files=temp_dir,
+        training_files=str(temp_dir),
         output=model_output_directory,
         force_training=request_payload.get(
             "force", request.args.get("force_training", False)
@@ -1207,7 +1445,7 @@ def _training_payload_from_json(request: Request) -> Dict[Text, Union[Text, bool
 def _validate_json_training_payload(rjs: Dict):
     if "config" not in rjs:
         raise ErrorResponse(
-            400,
+            HTTPStatus.BAD_REQUEST,
             "BadRequest",
             "The training request is missing the required key `config`.",
             {"parameter": "config", "in": "body"},
@@ -1215,7 +1453,7 @@ def _validate_json_training_payload(rjs: Dict):
 
     if "nlu" not in rjs and "stories" not in rjs:
         raise ErrorResponse(
-            400,
+            HTTPStatus.BAD_REQUEST,
             "BadRequest",
             "To train a Rasa model you need to specify at least one type of "
             "training data. Add `nlu` and/or `stories` to the request.",
@@ -1224,7 +1462,7 @@ def _validate_json_training_payload(rjs: Dict):
 
     if "stories" in rjs and "domain" not in rjs:
         raise ErrorResponse(
-            400,
+            HTTPStatus.BAD_REQUEST,
             "BadRequest",
             "To train a Rasa model with story training data, you also need to "
             "specify the `domain`.",
@@ -1240,42 +1478,36 @@ def _validate_json_training_payload(rjs: Dict):
         )
 
 
-def _training_payload_from_yaml(request: Request) -> Dict[Text, Union[Text, bool]]:
+def _training_payload_from_yaml(
+    request: Request, temp_dir: Path
+) -> Dict[Text, Union[Text, bool]]:
     logger.debug("Extracting YAML training data from request body.")
 
     decoded = request.body.decode(rasa.shared.utils.io.DEFAULT_ENCODING)
     _validate_yaml_training_payload(decoded)
 
-    temp_dir = tempfile.mkdtemp()
-    training_data = Path(temp_dir) / "data.yml"
+    training_data = temp_dir / "data.yml"
     rasa.shared.utils.io.write_text_file(decoded, training_data)
 
-    model_output_directory = _model_output_directory(
-        request.args.get("save_to_default_model_directory", True)
-    )
+    model_output_directory = str(temp_dir)
+    if request.args.get("save_to_default_model_directory", True):
+        model_output_directory = DEFAULT_MODELS_PATH
 
     return dict(
         domain=str(training_data),
         config=str(training_data),
-        training_files=temp_dir,
+        training_files=str(temp_dir),
         output=model_output_directory,
         force_training=request.args.get("force_training", False),
     )
 
 
-def _model_output_directory(save_to_default_model_directory: bool) -> Text:
-    if save_to_default_model_directory:
-        return DEFAULT_MODELS_PATH
-
-    return tempfile.gettempdir()
-
-
 def _validate_yaml_training_payload(yaml_text: Text) -> None:
     try:
         RasaYAMLReader().validate(yaml_text)
     except Exception as e:
         raise ErrorResponse(
-            400,
+            HTTPStatus.BAD_REQUEST,
             "BadRequest",
             f"The request body does not contain valid YAML. Error: {e}",
             help_url=DOCS_URL_TRAINING_DATA,
